import bpy
import bmesh
from mathutils import Vector, Matrix, Euler, Quaternion
import mathutils
from bpy.props import (
        IntProperty,
        FloatVectorProperty,
        EnumProperty,
        BoolProperty,
        FloatProperty
        )

_valid_types = ['MESH', 'CURVE', 'FONT']
#
#Generate a lattice and fit it to a selected object.
#
def fit_lattice(self, context, lattice_rotation):
    """fit a lattice to the selected object"""

    obj = context.active_object
    collections = obj.users_collection if len(obj.users_collection) else [context.collection]

    # check for edit mode...
    vg = None
    if context.active_object.mode == 'EDIT':
        bpy.ops.object.mode_set(mode='OBJECT')
        vertex_indices = [v.index for v in obj.data.vertices if v.select]
        vg = obj.vertex_groups.new(name="fit Lattice group")
        vg.add( vertex_indices, 1.0, 'REPLACE' )
        

    selected_objects = context.selected_objects
    lattice_data = bpy.data.lattices.new(obj.name + ' Lattice')
    lattice_object = bpy.data.objects.new(obj.name + ' Lattice', lattice_data)
    for collection in collections:
        collection.objects.link(lattice_object)

    lattice_object.location=self.obj_location

    # lattice_object = context.object
    dimensions = []
    dimensions.append(self.obj_dimensions[0] if self.obj_dimensions[0] else 0.001)
    dimensions.append(self.obj_dimensions[1] if self.obj_dimensions[1] else 0.001)
    dimensions.append(self.obj_dimensions[2] if self.obj_dimensions[2] else 0.001)

    lattice_object.dimensions=Vector(dimensions)
    lattice_data.points_u = self.points_u
    lattice_data.points_v = self.points_v
    lattice_data.points_w = self.points_w

    lattice_data.interpolation_type_u = self.interpolation_type
    lattice_data.interpolation_type_v = self.interpolation_type
    lattice_data.interpolation_type_w = self.interpolation_type
    
    lattice_object.rotation_euler = lattice_rotation.copy()
        
    modifier_added = False

    global _valid_types
    for target in selected_objects:
        if target.type in _valid_types:
            modifier = target.modifiers.new(name="Lattice", type='LATTICE')
            modifier.object = lattice_object
            if vg is not None:
                modifier.vertex_group = vg.name

            target.select_set(False)
            lattice_object.fit_lattice_object_refs.add().fit_lattice_object_ref = target
            modifier_added = True

    if modifier_added:
        lattice_object.select_set(True)
        context.view_layer.objects.active = lattice_object
    else:
        bpy.data.objects.remove(lattice_object)
        bpy.data.lattices.remove(lattice_data)


#
#Generate a mesh and fit it to a selected object with an optional bind.
#
def fit_mesh(self, context, lattice_rotation):
    """fit a lattice to the selected object"""

    obj = context.active_object
    collections = obj.users_collection if len(obj.users_collection) else [context.collection]

    # check for edit mode...
    vg = None
    if context.active_object.mode == 'EDIT':
        bpy.ops.object.mode_set(mode='OBJECT')
        vertex_indices = [v.index for v in obj.data.vertices if v.select]
        vg = obj.vertex_groups.new(name="fit mesh group")
        vg.add( vertex_indices, 1.0, 'REPLACE' )
        

    selected_objects = context.selected_objects
    obj_data = bpy.data.meshes.new(obj.name + ' Mesh')
    obj_object = bpy.data.objects.new(obj.name + ' Mesh', obj_data)
    for collection in collections:
        collection.objects.link(obj_object)

    obj_object.location=self.obj_location

    obj_object.display_type = 'WIRE'


    # Create a bmesh from the mesh, this allows for advanced mesh manipulation
    bm = bmesh.new()

    # Create a cube in bmesh
    # Create scaling matrices for each axis
    scale_matrix_x = mathutils.Matrix.Scale(self.obj_dimensions[0], 4, (1, 0, 0)) * self.mesh_scale  # Scale X
    scale_matrix_y = mathutils.Matrix.Scale(self.obj_dimensions[1], 4, (0, 1, 0)) * self.mesh_scale # Scale Y
    scale_matrix_z = mathutils.Matrix.Scale(self.obj_dimensions[2], 4, (0, 0, 1)) * self.mesh_scale # Scale Z

    # Combine the scaling matrices into one
    combined_scale_matrix = scale_matrix_x @ scale_matrix_y @ scale_matrix_z


    bmesh.ops.create_cube(bm, size=1.0, matrix=combined_scale_matrix)


    # Subdivide mesh
    edges_to_subdivide = []

    for e in bm.edges:
        is_valid_faces = len([f for f in e.link_faces if abs(f.normal.y) or abs(f.normal.z)]) == 2
        if is_valid_faces:
            edges_to_subdivide.append(e)

    if self.points_u-2 > 0:
        bmesh.ops.subdivide_edges(bm, edges=edges_to_subdivide, cuts=self.points_u-2, use_grid_fill=True)


    edges_to_subdivide = []

    for e in bm.edges:
        is_valid_faces = len([f for f in e.link_faces if abs(f.normal.x) or abs(f.normal.z)]) == 2
        if is_valid_faces:
            edges_to_subdivide.append(e)

    if self.points_v-2 > 0:
        bmesh.ops.subdivide_edges(bm, edges=edges_to_subdivide, cuts=self.points_v-2, use_grid_fill=True)

    edges_to_subdivide = []

    for e in bm.edges:
        is_valid_faces = len([f for f in e.link_faces if abs(f.normal.x) or abs(f.normal.y)]) == 2
        if is_valid_faces:
            edges_to_subdivide.append(e)

    if self.points_w-2 > 0:
        bmesh.ops.subdivide_edges(bm, edges=edges_to_subdivide, cuts=self.points_w-2, use_grid_fill=True)

    # Update the mesh with the new bmesh data
    bm.to_mesh(obj_data)
    bm.free() 
    
    obj_object.rotation_euler = lattice_rotation.copy()
        
    modifier_added = False

    global _valid_types
    for target in selected_objects:
        if target.type in _valid_types:
            modifier = target.modifiers.new(name="MeshDeform", type='MESH_DEFORM')
            modifier.object = obj_object
            if vg is not None:
                modifier.vertex_group = vg.name

            target.select_set(False)
            obj_object.fit_lattice_object_refs.add().fit_lattice_object_ref = target
            
            if self.do_mesh_bind:
                old_active = context.view_layer.objects.active
                try:
                    context.view_layer.objects.active = target
                    bpy.ops.object.meshdeform_bind(modifier=modifier.name)
                    context.area.tag_redraw()
                finally:
                    context.view_layer.objects.active = old_active


            
            modifier_added = True

    if modifier_added:
        obj_object.select_set(True)
        context.view_layer.objects.active = obj_object
    else:
        bpy.data.objects.remove(obj_object)
        bpy.data.lattices.remove(obj_data)




def get_location_and_bounds(vertices):
    co_final = vertices[0]
    x,y,z =  co_final
    min_x = x
    max_x = x
    min_y = y
    max_y = y
    min_z = z
    max_z = z
    for i in range(1, len(vertices)):
            v = vertices[i]
            x,y,z = v
            min_x = min(x, min_x)
            max_x = max(x, max_x)
            min_y = min(y, min_y)
            max_y = max(y, max_y)
            min_z = min(z, min_z)
            max_z = max(z, max_z)

    dim_x = max_x - min_x
    dim_y = max_y - min_y
    dim_z = max_z - min_z

    av_x = (max_x + min_x) / float(2)
    av_y = (max_y + min_y) / float(2)
    av_z = (max_z + min_z) / float(2)

    obj_dimensions = (dim_x, dim_y, dim_z)
    obj_location = (av_x, av_y, av_z)

    return obj_dimensions, obj_location


#
# Class that performs a fit lattice operation.
#
class MESH_OT_fit_lattice(bpy.types.Operator):
    """Fit Lattice Operator"""
    bl_idname = "mesh.fit_lattice"
    bl_label = "Fit Lattice"
    bl_options = {'REGISTER', 'UNDO', 'PRESET'}


    points_u : IntProperty (
            name="U",
            description="Point in U Direction (can't be changed when there are shape keys)",
            min=1,
            default=2,
    )

    points_v : IntProperty (
            name="V",
            description="Point in V Direction (can't be changed when there are shape keys)",
            min=1,
            default=2,
    )

    points_w : IntProperty (
            name="W",
            description="Point in W Direction (can't be changed when there are shape keys)",
            min=1,
            default=2,
    )

    interpolation_type : EnumProperty(items=(('KEY_LINEAR', 'Linear', ''),
                                            ('KEY_CARDINAL', 'Cardinal', ''),
                                            ('KEY_CATMULL_ROM', 'Catmull-Rom', ''),
                                            ('KEY_BSPLINE', 'BSpline', '')),
                                        name = "Interpolation Type", default='KEY_BSPLINE')

    lattice_rotation : FloatVectorProperty(
            name="Rotation",
            subtype='EULER'
            )
    

    fit_mode : EnumProperty(items=(('LATTICE', 'Lattice', ''),
                                            ('MESH', 'Mesh', '')),)
    
    do_mesh_bind : BoolProperty (
        name="Bind to Object",
        description="Perform a Bind of the Mesh Deform modifier created for this object.\nNOTE: this can make Blender look like it freezes, expect a delay in processing time",
        default = False,
        options={'SKIP_SAVE'}
    )

    mesh_scale : FloatProperty(
            name="Scale",
            description="Scale multiplier for mesh (can be helpful for binding boundary vertices)",
            min=0.000,
            default=1,
            precision=3,
            step=1
            )
    

    obj_dimensions = (0,0,0)
    obj_location = (0,0,0)


    def draw(self, context):
        layout = self.layout

        col = layout.column()

        col.separator()

        col.row().prop(self, 'fit_mode', expand=True)

        col.separator()

        col.prop(self, "points_u", text="U:")
        col.prop(self, "points_v", text="V:")
        col.prop(self, "points_w", text="W:")

        col.separator()

        col.prop(self, "lattice_rotation", text="Rotation")

        col.separator()
        if self.fit_mode == 'LATTICE':
            col.label(text="Interpolation Type:")
            col.prop(self, "interpolation_type", text="")
        else:
            col.prop(self, 'mesh_scale', text="Scale Mesh")
            col.separator()
            btn_text = "Bind (Expect a Delay)" if not self.do_mesh_bind else "Unbind"
            icon_id = "SORTTIME" if not self.do_mesh_bind else "BLANK1"
            col.prop(self, 'do_mesh_bind', text=btn_text, icon=icon_id, toggle=True)


    @classmethod
    def poll(cls, context):
        #must be in edit and face select mode.
        return len(context.selected_objects) and \
            context.active_object and \
                (context.active_object.mode == 'OBJECT' or context.active_object.mode == 'EDIT')

    def invoke(self, context, event):
        return self.execute(context)


    def execute(self, context):


        cos = []
        global _valid_types
        for obj in context.selected_objects:

            if obj.type not in _valid_types:
                continue

            obj_eval = obj.evaluated_get(bpy.context.evaluated_depsgraph_get())
            if obj.type == 'MESH':
                if obj.mode == 'EDIT':
                    bm = bmesh.from_edit_mesh(obj.data)
                    cos.extend([obj_eval.matrix_world @ v.co for v in bm.verts if v.select])
                elif obj.mode == 'OBJECT':
                    bm = bmesh.new()
                    bm.from_mesh(obj_eval.data)
                    try:
                        cos.extend([obj_eval.matrix_world @ v.co for v in bm.verts])
                    finally:
                        bm.free()

            elif obj.type in {'CURVE', 'FONT'}:
                if obj.mode == 'EDIT':
                    self.report({'ERROR'}, F'Object type {obj.type} not supported in Edit mode.')
                    return {'CANCELLED'}
                elif obj.mode == 'OBJECT':
                    mesh = bpy.data.meshes.new_from_object(obj_eval)
                    try:
                        cos.extend([obj_eval.matrix_world @ v.co for v in mesh.vertices])
                    finally:
                        bpy.data.meshes.remove(mesh)
                


        if not cos:
            return {'CANCELLED'}

        x_coords = [co.x for co in cos]
        y_coords = [co.y for co in cos]
        z_coords = [co.z for co in cos]

        center = ((max(x_coords) + min(x_coords)) / 2,
                                (max(y_coords) + min(y_coords)) / 2,
                                (max(z_coords) + min(z_coords)) / 2)


        up = Vector((0,0,1))
        down = Vector((0,0,-1))
        left = Vector((1,0,0))
        right = Vector((-1,0,0))
        front = Vector((0,1,0))
        back = Vector((0,-1,0))

        lattice_rotation = self.lattice_rotation.copy()

        quaternions = [context.active_object.rotation_euler.to_quaternion()]

        # Average the quaternions
        avg_quat = Quaternion((0, 0, 0, 0))
        for quat in quaternions:
            avg_quat += quat
        avg_quat.normalize()

        average_euler = avg_quat.to_euler('XYZ')

        lattice_rotation.rotate(average_euler)


        up.rotate(lattice_rotation)       
        down.rotate(lattice_rotation)
        left.rotate(lattice_rotation)
        right.rotate(lattice_rotation)
        front.rotate(lattice_rotation)
        back.rotate(lattice_rotation)

        def max_dist(co):
            return mathutils.geometry.distance_point_to_plane(co, Vector(center), norm)
        
        norm = up
        up_co = max(cos, key=max_dist)
        up_dist = max_dist(up_co)
        up_co = Vector(center)  + (norm * up_dist)

        norm = down
        down_co = max(cos, key=max_dist)
        down_dist = max_dist(down_co)
        down_co = Vector(center)  + (norm * down_dist)

        norm = left
        left_co = max(cos, key=max_dist)
        left_dist = max_dist(left_co)
        left_co = Vector(center)  + (norm * left_dist)

        norm = right
        right_co = max(cos, key=max_dist)
        right_dist = max_dist(right_co)
        right_co = Vector(center)  + (norm * right_dist)

        norm = front
        front_co = max(cos, key=max_dist)
        front_dist = max_dist(front_co)
        front_co = Vector(center)  + (norm * front_dist)

        norm = back
        back_co = max(cos, key=max_dist)
        back_dist = max_dist(back_co)
        back_co = Vector(center)  + (norm * back_dist)

        # find the center by first finding intersecting with the top plan, facing down, and then the same with the down plane.......

        pt1, vec1 = mathutils.geometry.intersect_plane_plane(up_co, down, left_co, right)
        pt2, vec2 = mathutils.geometry.intersect_plane_plane(up_co, down, right_co, left)
        pt3, vec3 = mathutils.geometry.intersect_plane_plane(up_co, down, front_co, back)
        pt4, vec4 = mathutils.geometry.intersect_plane_plane(up_co, down, back_co, front)

        point1, _ = mathutils.geometry.intersect_line_line(pt1, pt1 + vec1, pt3, pt3 + vec3)
        point2, _ = mathutils.geometry.intersect_line_line(pt1, pt1 + vec1, pt4, pt4 + vec4)
        point3, _ = mathutils.geometry.intersect_line_line(pt2, pt2 + vec2, pt3, pt3 + vec3)
        point4, _ = mathutils.geometry.intersect_line_line(pt2, pt2 + vec2, pt4, pt4 + vec4)


        pt5, vec5 = mathutils.geometry.intersect_plane_plane(down_co, up, left_co, right)
        pt6, vec6 = mathutils.geometry.intersect_plane_plane(down_co, up, right_co, left)
        pt7, vec7 = mathutils.geometry.intersect_plane_plane(down_co, up, front_co, back)
        pt8, vec8 = mathutils.geometry.intersect_plane_plane(down_co, up, back_co, front)

        point5, _ = mathutils.geometry.intersect_line_line(pt5, pt5 + vec5, pt7, pt7 + vec7)
        point6, _ = mathutils.geometry.intersect_line_line(pt5, pt5 + vec5, pt8, pt8 + vec8)
        point7, _ = mathutils.geometry.intersect_line_line(pt6, pt6 + vec6, pt7, pt7 + vec7)
        point8, _ = mathutils.geometry.intersect_line_line(pt6, pt6 + vec6, pt8, pt8 + vec8)

        all_coords = [point1, point2, point3, point4, point5, point6, point7, point8]

        center = sum(all_coords, Vector()) / len(all_coords)

        self.obj_dimensions = (left_dist + right_dist, front_dist + back_dist, up_dist + down_dist)
        self.obj_location = Vector(center)

        
        #perform the mapping
        if self.fit_mode == 'LATTICE':
            fit_lattice(self, context, lattice_rotation)
        else:
            fit_mesh(self, context, lattice_rotation)

        return {'FINISHED'}


#
# Class that removes a fit lattice and associated modifiers/vertex groups.
#
class MESH_OT_remove_fit_lattice(bpy.types.Operator):
    """Fit Lattice Remove Operator"""
    bl_idname = "mesh.fit_lattice_remove"
    bl_label = "Remove Fit Lattice"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return context.active_object and \
                context.active_object.type == 'LATTICE'

    def invoke(self, context, event):
        return self.execute(context)


    def execute(self, context):
        
        # get the ref assigned to the lattice.
        objs_using_lat = context.active_object.fit_lattice_object_refs

        for obj_using_lat_ref in objs_using_lat:
            obj_using_lat = obj_using_lat_ref.fit_lattice_object_ref
            for mod in obj_using_lat.modifiers:
                if (mod.type == 'LATTICE' or mod.type == 'MESH_DEFORM') and mod.object == context.active_object:
                    # get the associated vertex group and remove it.
                    vertex_group_to_delete = mod.vertex_group
                    if vertex_group_to_delete in obj_using_lat.vertex_groups:
                        obj_using_lat.vertex_groups.remove(obj_using_lat.vertex_groups[vertex_group_to_delete])
                    #remove the modifier
                    obj_using_lat.modifiers.remove(mod)

        # remove the lattice object.
        try:
            old_mesh = context.active_object.data
            bpy.data.objects.remove(context.active_object, do_unlink=True)
            if old_mesh.name in bpy.data.lattices:
                bpy.data.lattices.remove(old_mesh, do_unlink=True)
            if old_mesh.name in bpy.data.meshes:
                bpy.data.meshes.remove(old_mesh, do_unlink=True)

        except:
            pass
        return {'FINISHED'}

def fit_lattice_menu_func(layout):
    """declare the menu function"""
    layout.operator(MESH_OT_fit_lattice.bl_idname, text="Fit Lattice")
    layout.operator(MESH_OT_remove_fit_lattice.bl_idname, text="Remove Fit Lattice")

def add_to_key_map():
    # handle the keymap
    wm = bpy.context.window_manager
    km1 = wm.keyconfigs.addon.keymaps.new(name = "Window",space_type='EMPTY', region_type='WINDOW')
    kmi1 = km1.keymap_items.new(MESH_OT_fit_lattice.bl_idname, 'L', 'PRESS', alt=True, ctrl=False, shift=False)
    addon_keymaps.append(km1)

# store keymaps here to access after registration
addon_keymaps = []
def remove_from_key_map():
    # handle the keymap
    wm = bpy.context.window_manager
    for km in addon_keymaps:
        wm.keyconfigs.addon.keymaps.remove(km)
    # clear the list
    addon_keymaps.clear()
